import asyncio
import csv
import psutil
import uuid
from typing import List
from pathlib import Path
from concurrent.futures import ThreadPoolExecutor
from hedra.logging import HedraLogger
from hedra.reporting.events.types.base_event import BaseEvent
from hedra.reporting.metric.metrics_set import MetricsSet
from .csv_config import CSVConfig
has_connector = True


class CSV:

    def __init__(self, config: CSVConfig) -> None:
        self.events_filepath = config.events_filepath
        self.metrics_filepath = config.metrics_filepath
        self._executor = ThreadPoolExecutor(max_workers=psutil.cpu_count(logical=False))
        self._loop = asyncio.get_event_loop()

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

        self._events_csv_writer = None
        self._metrics_csv_writer = None
        self._stage_metrics_csv_writer = None
        self._errors_csv_writer = None
        self._custom_metrics_csv_writers = {}

    async def connect(self):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Skipping connect')

    async def submit_events(self, events: List[BaseEvent]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Events to file - {self.events_filepath}')
        with open(self.events_filepath, 'a') as events_file:

            for event in events:
                if self._events_csv_writer is None:
                    self._events_csv_writer = csv.DictWriter(events_file, fieldnames=event.fields)
                    
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Writing headers to file - {self.metrics_filepath} - {", ".join(event.fields)}')

                    await self._loop.run_in_executor(
                        self._executor,
                        self._events_csv_writer.writeheader
                    )

                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Wrote headers to file - {self.metrics_filepath} - {", ".join(event.fields)}')

                await self._loop.run_in_executor(
                    self._executor,
                    self._events_csv_writer.writerow,
                    event.record
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saved Events to file - {self.events_filepath}')

    async def submit_common(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Shared Metrics to file - {self.metrics_filepath}')

        headers = [
            'name',
            'stage',
            'group',
            'total',
            'succeeded',
            'failed'
        ]

        base_filepath = Path(self.metrics_filepath).parent
        with open(f'{base_filepath}/stage_metrics.csv', 'a') as shared_metrics_file:

            if self._stage_metrics_csv_writer is None:
                self._stage_metrics_csv_writer = csv.DictWriter(shared_metrics_file, fieldnames=headers)


                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Writing headers to file - {self.metrics_filepath} - {", ".join(headers)}')

                await self._loop.run_in_executor(
                    self._executor,
                    self._stage_metrics_csv_writer.writeheader
                )

                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Wrote headers to file - {self.metrics_filepath} - {", ".join(headers)}')

            for metrics_set in metrics_sets:
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

                await self._loop.run_in_executor(
                    self._executor,
                    self._stage_metrics_csv_writer.writerow,
                    {
                        'name': metrics_set.name,
                        'stage': metrics_set.stage,
                        'group': 'common',
                        **metrics_set.common_stats
                    }
                )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saved Shared Metrics to file - {self.metrics_filepath}')

    async def submit_metrics(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Metrics to file - {self.metrics_filepath}')

        with open(self.metrics_filepath, 'a') as metrics_file:

            for metrics_set in metrics:
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

                if self._metrics_csv_writer is None:

                    headers = [
                        *metrics_set.fields,
                        'group'
                    ]

                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Writing headers to file - {self.metrics_filepath} - {", ".join(headers)}')

                    self._metrics_csv_writer = csv.DictWriter(metrics_file, fieldnames=headers)

                    await self._loop.run_in_executor(
                        self._executor,
                        self._metrics_csv_writer.writeheader
                    )

                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Wrote headers to file - {self.metrics_filepath} - {", ".join(headers)}')

                for group_name, group in metrics_set.groups.items():
                    await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Group - {group_name}:{group.metrics_group_id}')

                    await self._loop.run_in_executor(
                        self._executor,
                        self._metrics_csv_writer.writerow,
                        {
                            **group.record,
                            'group': group_name
                        }
                    )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saved Metrics to file - {self.metrics_filepath}')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Custom Metrics to file - {self.metrics_filepath}')

        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')
                
            for custom_group_name, group in metrics_set.custom_metrics.items():
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Group - {custom_group_name}')

                base_filepath = Path(self.metrics_filepath).parent
                with open(f'{base_filepath}/{custom_group_name}.csv', 'a') as custom_metrics_file:

                    headers = [
                        'name',
                        'stage',
                        'group',
                        *list(group.keys())
                    ]
                    
                    if self._custom_metrics_csv_writers.get(custom_group_name) is None:

                        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Writing headers to file - {self.metrics_filepath} - {", ".join(headers)}')

                        custom_group_csv_writer = csv.DictWriter(custom_metrics_file, fieldnames=headers)
                        self._custom_metrics_csv_writers[custom_group_name] = custom_group_csv_writer

                        await self._loop.run_in_executor(
                            self._executor,
                            custom_group_csv_writer.writeheader
                        )

                        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Wrote headers to file - {self.metrics_filepath} - {", ".join(headers)}')


                    await self._loop.run_in_executor(
                        self._executor,
                        self._custom_metrics_csv_writers[custom_group_name].writerow,
                        {
                            'name': metrics_set.name,
                            'stage': metrics_set.stage,
                            'group': custom_group_name,
                            **group
                        }
                    )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saved Custom Metrics to file - {self.metrics_filepath}')

    async def submit_errors(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saving Error Metrics to file - {self.metrics_filepath}')

        error_csv_headers = [
            'name',
            'stage',
            'error_message',
            'error_count'
        ]

        base_filepath = Path(self.metrics_filepath).parent
        with open (f'{base_filepath}/stage_errors.csv', 'a') as errors_file:
            
            if self._errors_csv_writer is None:

                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Writing headers to file - {self.metrics_filepath} - {", ".join(error_csv_headers)}')
                error_csv_writer = csv.DictWriter(errors_file, fieldnames=error_csv_headers)

                await self._loop.run_in_executor(
                    self._executor,
                    error_csv_writer.writeheader
                )

                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Wrote headers to file - {self.metrics_filepath} - {", ".join(error_csv_headers)}')

            for metrics_set in metrics_sets:
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Error Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

                for error in metrics_set.errors:
                    await self._loop.run_in_executor(
                        self._executor,
                        error_csv_writer.writerow,
                        {
                            'name': metrics_set.name,
                            'stage': metrics_set.stage,
                            'error_message': error.get('message'),
                            'error_count': error.get('count')
                        }
                    )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Saved Error Metrics to file - {self.metrics_filepath}')
                
    async def close(self):
        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')





from google.protobuf.json_format import MessageToDict
import grpc
from hedra.runners.utils.connect_timeout import connect_or_return_none
from hedra.runners.leader_services.proto import (
    DistributedServerStub,
    LeaderRegisterRequest,
    PollLeaderRequest,
    LeaderJobRequest
)


class LeaderService:


    def __init__(self, service_address, service_id, host=False) -> None:
        self.service_address = service_address
        self.service_id = service_id
        self.channel = grpc.aio.insecure_channel(service_address)
        self.service = DistributedServerStub(self.channel)
        self.active = True
        self.host = host

    async def is_registered(self, registered_services):
        return self.service_address in registered_services.registered

    @connect_or_return_none(wait_buffer=1, timeout_threshold=3)
    async def register(self, service_address, service_id):
        request = LeaderRegisterRequest(
                host_address=service_address,
                host_id=str(service_id)
            )

        return await self.service.RegisterLeader(request)

    @connect_or_return_none(wait_buffer=1, timeout_threshold=3)
    async def poll(self):
        request = PollLeaderRequest(
            host_address=self.service_address,
            host_id=str(self.service_id)
        )
        return await self.service.PollLeader(request)

    @connect_or_return_none(wait_buffer=5, timeout_threshold=15)
    async def share_current_job_config(self, job):

        request = LeaderJobRequest(
            host_address=self.service_address,
            host_id=str(self.service_id),
            job_id=job.job_id
        )

        request.job_config.update(job.config)
        request.pipeline_status.update(job.pipeline.reported_completions)

        response = await self.service.ShareCurrentJob(request) 
        updated_pipeline_status = MessageToDict(response.pipeline_status)

        return updated_pipeline_status

           



import os
from hedra.projects.generation import PluginGenerator
from hedra.cli.exceptions.plugin.create import InvalidPluginType
from hedra.logging import HedraLogger
from hedra.logging import (
    HedraLogger,
    LoggerTypes,
    logging_manager
)


def create_plugin(plugin_type: str, path: str, log_level: str):

    logging_manager.disable(
        LoggerTypes.HEDRA, 
        LoggerTypes.DISTRIBUTED,
        LoggerTypes.FILESYSTEM,
        LoggerTypes.DISTRIBUTED_FILESYSTEM
    )

    logging_manager.update_log_level(log_level)

    logger = HedraLogger()
    logger.initialize()
    logging_manager.logfiles_directory = os.getcwd()

    logger['console'].sync.info(f'Creating new - {plugin_type} - plugin at - {path}.')

    generator = PluginGenerator()
    generated_plugin_data = None

    if plugin_type in generator.generator_types:
        generated_plugin_data = generator.generate_plugin(plugin_type)

    else:
        raise InvalidPluginType(
            plugin_type,
            list(generator.generator_types.keys())
        )

    logger['console'].sync.info('Saving template.')

    with open(path, 'w') as generated_plugin:
        generated_plugin.write(f'{generated_plugin_data}\n')

    logger['console'].sync.info('\nPlugin generated!\n')
    


import asyncio
import traceback
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs.stages.types.stage_states import StageStates
from hedra.core.graphs.stages.types.stage_types import StageTypes
from hedra.core.graphs.transitions.exceptions import (
    StageExecutionError,
    StageTimeoutError
)
from hedra.core.graphs.stages.validate.exceptions import (
    HookValidationError,
    ReservedMethodError,
    MissingReservedMethodError
)
from hedra.logging import HedraLogger


async def validate_to_setup_transition(current_stage: Stage, next_stage: Stage):

    logger = HedraLogger()
    logger.initialize()

    try:

        if current_stage.state == StageStates.INITIALIZED:

            await logger.spinner.system.debug(f'{current_stage.metadata_string} - Executing transition from {current_stage.name} to {next_stage.name}')
            await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - Executing transition from {current_stage.name} to {next_stage.name}')

            current_stage.state = StageStates.VALIDATING
            current_stage.stages = current_stage.context.stages
            
            if current_stage.timeout:
                await asyncio.wait_for(current_stage.run(), timeout=current_stage.timeout)

            else:
                await current_stage.run()
            
            current_stage.state = StageStates.VALIDATED

            await logger.spinner.system.debug(f'{current_stage.metadata_string} - Completed transition from {current_stage.name} to {next_stage.name}')
            await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - Completed transition from {current_stage.name} to {next_stage.name}')

        else:
            await logger.spinner.system.debug(f'{current_stage.metadata_string} - Skipping transition from {current_stage.name} to {next_stage.name}')
            await logger.filesystem.aio['hedra.core'].debug(f'{current_stage.metadata_string} - Skipping transition from {current_stage.name} to {next_stage.name}')

        next_stage.state = StageStates.VALIDATED

        next_stage.context = current_stage.context

    except MissingReservedMethodError as missing_reserved_method_error:
        return missing_reserved_method_error, StageTypes.ERROR

    except ReservedMethodError as reserved_method_error:
        return reserved_method_error, StageTypes.ERROR

    except HookValidationError as hook_validation_error:
        return hook_validation_error, StageTypes.ERROR
            
    except asyncio.TimeoutError:
        return StageTimeoutError(current_stage), StageTypes.ERROR

    except Exception as stage_execution_error:
        print(traceback.format_exc())
        return StageExecutionError(current_stage, next_stage, str(stage_execution_error)), StageTypes.ERROR
    
    else:
        return None, StageTypes.SETUP


async def validate_to_wait_transition(current_stage: Stage, next_stage: Stage):

    try:

        if current_stage.state == StageStates.INITIALIZED:
            current_stage.state = StageStates.VALIDATING
            current_stage.stages = current_stage.context.stages
            
            if current_stage.timeout:
                await asyncio.wait_for(current_stage.run(), timeout=current_stage.timeout)

            else:
                await current_stage.run()
            
            current_stage.state = StageStates.VALIDATED

        next_stage.state = StageStates.VALIDATED

        next_stage.context = current_stage.context

    except MissingReservedMethodError as missing_reserved_method_error:
        return missing_reserved_method_error, StageTypes.ERROR

    except ReservedMethodError as reserved_method_error:
        return reserved_method_error, StageTypes.ERROR

    except HookValidationError as hook_validation_error:
        return hook_validation_error, StageTypes.ERROR
            
    except asyncio.TimeoutError:
        return StageTimeoutError(current_stage), StageTypes.ERROR

    except Exception as stage_execution_error:
        return StageExecutionError(current_stage, next_stage, str(stage_execution_error)), StageTypes.ERROR
    
    else:
        return None, StageTypes.WAIT


from typing import Any, List
try:
    from prometheus_client import (
        Info,
        Summary,
        Counter,
        Gauge,
        Histogram,
        Enum
    )

except Exception:
    Info = None
    Summary = None
    Counter = None
    Gauge = None
    Histogram = None
    Enum = None
    pass


class PrometheusMetric:

    types = {
            'info': Info,
            'summary': Summary,
            'count': Counter,
            'gauge': Gauge,
            'histogram': Histogram,
            'enum': Enum
        }

    def __init__(
        self,
        metric_name: str, 
        metric_type: str, 
        metric_description: str=None,
        metric_states: List[Any]= None,
        metric_labels: List[str]=[],
        metric_namespace: str = None,
        registry=None
    ):
        self.name = metric_name
        self.type = metric_type
        self.description = metric_description
        self.states = metric_states
        self.label_names = metric_labels
        self.namespace = metric_namespace
        self.registry = registry
        self.prometheus_object = None

    def create_metric(self):
        
        if self.type == 'enum':
            self.prometheus_object = Enum(
                self.name,
                self.description,
                labelnames=(label for label in self.label_names),
                states=self.states,
                registry=self.registry,
                namespace=self.namespace
            )

        elif self.type == 'gauge':
            self.prometheus_object = Gauge(
                self.name,
                self.description,
                labelnames=(label for label in self.label_names),
                registry=self.registry,
                namespace=self.namespace
            )

        elif self.type == 'summary':
            self.prometheus_object = Summary(
                self.name,
                self.description,
                labelnames=(label for label in self.label_names),
                registry=self.registry,
                namespace=self.namespace
            )

        elif self.type == 'count':
            self.prometheus_object = Counter(
                self.name,
                self.description,
                labelnames=(label for label in self.label_names),
                registry=self.registry,
                namespace=self.namespace
            )

        elif self.type == 'histogram':
            self.prometheus_object = Histogram(
                self.name,
                self.description,
                labelnames=(label for label in self.label_names),
                registry=self.registry,
                namespace=self.namespace
            )

        elif self.type == 'info':
            self.prometheus_object = Info(
                self.name,
                self.description,
                labelnames=(label for label in self.label_names),
                registry=self.registry,
                namespace=self.namespace
            )

        self.prometheus_object._labelnames = self.label_names

        self.generators = {
            'info': self._update_info,
            'enum': self._update_enum,
            'histogram': self._update_histogram_or_summary,
            'summary': self._update_histogram_or_summary,
            'count': self._update_count,
            'gauge': self._update_gauge,
        }

    def update(self, value=None, labels=None, options=None) -> None:
        if labels is None:
            labels = {label: '' for label in self.label_names}

            self.prometheus_object.labels(**labels)

        self.generators.get(self.type)(value, labels, options)

    def _update_info(self, value, labels, options) -> None:
        if value is None:
            value = 'N/A'

        self.prometheus_object.info(value)

    def _update_enum(self, value, labels, options) -> None:
        if value is None:
            value = 'N/A'

        self.prometheus_object.state(value)

    def _update_histogram_or_summary(self, value, labels, options) -> None:
        if value is None:
            value = 0

        self.prometheus_object.observe(value)

    def _update_count(self, value, labels, options) -> None:
        self.prometheus_object.inc()

    def _update_gauge(self, value, labels, options) -> None:
        if value is None:
            value = 1

        if options is None:
            options = {}

        if options.get('update_type') == 'inc':
            self.prometheus_object.inc(amount=value)
        elif options.get('update_type') == 'dec':
            self.prometheus_object.dec(amount=value)
        elif options.get('update_type') == 'set_function':
            self.prometheus_object.set_function(
                options.get('update_function')(value)
            )
        else:
            self.prometheus_object.set(value)



from hedra.core.graphs.stages import (
    Submit
)

from hedra.reporting.types import (
    TelegrafConfig
)


class SubmitTelegrafResultsStage(Submit):
    config=TelegrafConfig(
        host='localhost',
        port=8094
    )



from __future__ import annotations
import inspect
from typing import Any, Generic, Optional, TypeVar

from pydantic import create_model

T = TypeVar('T')


class ReporterConfig(Generic[T]):
    reporter_type: Optional[str]

    def __init__(self, **data: Any) -> None:
        super().__init__()

        attributes = inspect.getmembers(
            self, 
            lambda attr: not(inspect.isroutine(attr))
        )

        instance_attributes = [
            attr for attr in attributes if not(
                attr[0].startswith('__') and attr[0].endswith('__')
            )
        ]

        base_attributes = set(dir(ReporterConfig))
        
        model_attributes = {}
        for attribute_name, attribute_value in instance_attributes:
            
            if attribute_name not in base_attributes:
                model_attributes[attribute_name] = attribute_value

        model_attributes.update(data)

        for attribute_name, attribute_value in data.items():
            setattr(self, attribute_name, attribute_value)

        validation_model = create_model(
            self.__class__.__name__,
            **{
                field_name: (
                    field_type, 
                    ...
                ) for field_name, field_type in self.__class__.__annotations__.items()
            }
        )
        
        validation_model(**model_attributes)


from hedra.core.engines.types.graphql import GraphQLResult
from .http_event import HTTPEvent


class GraphQLEvent(HTTPEvent):

    def __init__(self, result: GraphQLResult) -> None:
        super(GraphQLEvent, self).__init__(result)


from typing import Optional

class RepoConfig:

    def __init__(self, 
        path: str, 
        uri: str, 
        branch: str='main',
        remote: str='origin', 
        sync_message: Optional[str]=None,
        username: Optional[str]=None, 
        password: Optional[str]=None,
        ignore_options: Optional[str]=None
    ) -> None:

        self.path = path
        self.uri = uri
        self.branch = branch
        self.remote = remote
        self.sync_message = sync_message
        self.username = username
        self.password = password
        self.ignore_options = [
            '__pycache__',
            '**/*.pyx'
        ]

        if isinstance(ignore_options, str) and len(ignore_options) > 0:
            self.ignore_options.extend(
                ignore_options.split(',')
            )




import asyncio
import os
import grpc
from hedra.tools.data_structures import AsyncList, AsyncDict
from hedra.tools.helpers import awaitable
from hedra.runners.leader_services.proto import (
    WorkerUpdateRequest,
    DistributedServerStub
)
from google.protobuf.json_format import (
    MessageToDict
)

from hedra.command_line import CommandLine
from hedra.runners.utils.connect_timeout import connect_with_retry

class UpdateService:

    def __init__(self, command_line=None):
        logger = Logger()
        if command_line is None:
            command_line = CommandLine()
            
        self.config = command_line.distributed_config
        self.session_logger = logger.generate_logger('hedra')
        leader_ips = command_line.distributed_config.get(
            'leader_ips',
            os.getenv('LEADER_IPS', 'localhost')
        )

        leader_ports = command_line.distributed_config.get(
            'leader_ports',
            os.getenv('LEADER_PORTS', "6669")
        )

        leader_addresses = command_line.distributed_config.get(
            'leader_addresses',
            os.getenv('LEADER_ADDRESSES')
        )

        self.leader_ips = {}
        self.leader_ports = {}

        if leader_addresses:
            leader_addresses = leader_addresses.split(",")
            self.leader_addresses = leader_addresses
            for leader_address in self.leader_addresses:
                leader_ip, leader_port = leader_address.split(':')
                self.leader_ips[leader_address] = leader_ip
                self.leader_ports[leader_address] = int(leader_port)

        else:
            leader_ips = leader_ips.split(",")
            leader_ports = [int(port) for port in leader_ports.split(",")]

            self.leader_addresses = []

            for leader_ip, leader_port in zip(leader_ips, leader_ports):
                leader_address = f'{leader_ip}:{leader_port}'
                self.leader_addresses.append(leader_address)

                self.leader_ips[leader_address] = leader_ip
                self.leader_ports[leader_address] = leader_port

        self.channels = {}
        self.services = {}

        for leader_address in self.leader_addresses:
            channel = grpc.aio.insecure_channel(leader_address)

            self.channels[leader_address] = channel
            self.services[leader_address] = DistributedServerStub(channel)

        self.services = AsyncDict(self.services)

        self.endpoint_configs = [
            {
                'name': 'get_worker_updates',
                'path': '/api/hedra/workers/updates',
                'methods': ['GET']
            }
        ]

        self.loop = asyncio.get_event_loop()

    def __iter__(self):
        for endpoint_config in self.endpoint_configs:
            yield endpoint_config, getattr(self, endpoint_config.get('name'))

    def get_worker_updates(self):
        update_response = self.loop.run_until_complete(self.GetUpdate())
        return {
            'data': update_response
        }, 200

    @connect_with_retry(timeout_threshold=15)
    async def GetUpdate(self):
        updates = AsyncList()

        async for service_address, service in self.services.iter_items():
            leader_ip = self.leader_ips[service_address]
            leader_port = self.leader_ports[service_address]

            update_request = WorkerUpdateRequest(
                leader_address=leader_ip,
                leader_port=leader_port
            )
            update_response = await service.GetUpdate(update_request)
            update_response = MessageToDict(update_response)
            
            await updates.append(update_response)

        return updates.data







from .base_event import BaseEvent


class TrailersReceived(BaseEvent):
    event_type='TRAILERS_RECEIVED'
    __slots__ = ('stream_id', 'headers', 'stream_ended', 'priority_updated')
    """
    The TrailersReceived event is fired whenever trailers are received on a
    stream. Trailers are a set of headers sent after the body of the
    request/response, and are used to provide information that wasn't known
    ahead of time (e.g. content-length). This event carries the HTTP header
    fields that form the trailers and the stream ID of the stream on which they
    were received.

    .. versionchanged:: 2.3.0
       Changed the type of ``headers`` to :class:`HeaderTuple
       <hpack:hpack.HeaderTuple>`. This has no effect on current users.

    .. versionchanged:: 2.4.0
       Added ``stream_ended`` and ``priority_updated`` properties.
    """
    def __init__(self):
        #: The Stream ID for the stream on which these trailers were received.
        self.stream_id = None

        #: The trailers themselves.
        self.headers = None

        #: Trailers always end streams. This property has the associated
        #: :class:`StreamEnded <h2.events.StreamEnded>` in it.
        #:
        #: .. versionadded:: 2.4.0
        self.stream_ended = None

        #: If the trailers also set associated priority information, the
        #: associated :class:`PriorityUpdated <h2.events.PriorityUpdated>`
        #: event will be available here.
        #:
        #: .. versionadded:: 2.4.0
        self.priority_updated = None

    def __repr__(self):
        return "<TrailersReceived stream_id:%s, headers:%s>" % (
            self.stream_id, self.headers
        )


import asyncio
from asyncio.tasks import wait
import os
import grpc

from hedra.runners.worker_services.bootstrap import BootstrapManager
from .leader_registry import LeaderRegistry

from hedra.runners.leader_services.proto import (
    WorkerRegisterRequest,
    WorkerDeregisterRequest,
    LeaderHeartbeatRequest
)
from hedra.runners.utils.connect_timeout import (
    connect_with_retry_async,
    connect_or_return_none
)


class LeaderServicesManager:

    def __init__(self, config, worker_ip, worker_port, worker_id) -> None:
        logger = Logger()
        self.session_logger = logger.generate_logger()
        self.worker_ip = worker_ip
        self.worker_port = worker_port
        self.worker_id = worker_id
        self.worker_address = f'{self.worker_ip}:{self.worker_port}'

        self.worker_server_port = config.distributed_config.get(
            'worker_server_port',
            os.getenv('WORKER_SERVER_PORT',  "6671")
        )

        self.worker_server_address = f'{self.worker_ip}:{self.worker_server_port}'


        self.leader_ips = config.distributed_config.get(
            'leader_ips',
            os.getenv('LEADER_IPS', 'localhost')
        )

        self.leader_ports = config.distributed_config.get(
            'leader_ports',
            os.getenv('LEADER_PORTS', "6669")
        )

        self.leader_request_timeout =config.distributed_config.get(
            'leader_request_timeout',
            30
        )

        leader_addresses = config.distributed_config.get(
            'leader_addresses',
            os.getenv('LEADER_ADDRESSES')
        )

        if leader_addresses:
            leader_addresses = leader_addresses.split(",")
            self.leader_addresses = leader_addresses

        else:
            self.leader_ips = self.leader_ips.split(",")
            self.leader_ports = [int(port) for port in self.leader_ports.split(",")]

            self.leader_addresses = []

            for leader_ip, leader_port in zip(self.leader_ips, self.leader_ports):
                self.leader_addresses.append(
                    f'{leader_ip}:{leader_port}'
                )

        leader_address_strings = ''.join([f'\n- {leader_address}' for leader_address in self.leader_addresses])
        
        self.session_logger.info(f'\nConnecting to leaders at addresses: {leader_address_strings}\n')

        self.registry = LeaderRegistry(self.leader_addresses)
        self.broadkast_manager = BootstrapManager(config)

    async def initialize(self):
        await self.registry.initialize()

    async def registered(self):
        return self.registry.leader_addresses

    async def add_new_leaders(self, new_leaders_queue):
        
        has_new_leaders = await new_leaders_queue.not_empty()
        if has_new_leaders:        
            await self.register_worker(leaders=new_leaders_queue)

        return new_leaders_queue

    async def register_worker(self, leaders=None):

        if leaders is None:
            if self.broadkast_manager.use_bootstrap:
                await self.broadkast_manager.client.discover()
                leaders = self.broadkast_manager.client.registry

            else:
                leaders = self.registry

        responses, _ = await asyncio.wait([
            self._register_worker(
                leaders[leader_address]
            ) for leader_address in leaders.leader_addresses
        ], timeout=self.leader_request_timeout)

        results = await asyncio.gather(*responses)
        
        for registration_response in results:
            await self.registry.register_leader(registration_response.host_address)

        self.session_logger.debug('Worker successfully registered.')

    @connect_with_retry_async(wait_buffer=5, timeout_threshold=15)
    async def _register_worker(self, leader):
        register_request = WorkerRegisterRequest(
            host_address=self.worker_address,
            host_id=str(self.worker_id),
            server_address=self.worker_server_address
        )

        return await leader.client.AddClient(register_request)

    async def deregister_worker(self, executor):
        
        self.session_logger.debug('Deregistring client...')

        await asyncio.wait([
            await self._deregister_worker(
                self.registry[leader_address], 
                executor.handler.reporter.fields
            ) for leader_address in self.registry.registered_leaders
        ], timeout=self.leader_request_timeout)

        self.session_logger.info('Closing down worker. Goodbye!')

    @connect_with_retry_async(wait_buffer=5, timeout_threshold=15)
    async def _deregister_worker(self, leader, reporter_fields):
        deregister_request = WorkerDeregisterRequest(
            host_address=self.worker_address,
            host_id=str(self.worker_id),
        )

        deregister_request.reporter_fields.extend(reporter_fields)

        await leader.client.RemoveClient(deregister_request)

        await self.registry.remove_leader(leader.address)

    async def check_leader_heartbeat(self):

        self.session_logger.debug('Getting leader heartbeats...')

        for leader in self.registry:
            heartbeat_response = await self._check_leader_heartbeat(leader)
            await self._update_registry(heartbeat_response,leader)
            
    @connect_or_return_none(wait_buffer=1, timeout_threshold=1)
    async def _check_leader_heartbeat(self, leader):
        heartbeat_request = LeaderHeartbeatRequest(
            host_address=self.worker_address,
            host_id=str(self.worker_id),
            status='OK'
        )

        return await leader.client.CheckLeaderHeartbeat(heartbeat_request)

    async def _update_registry(self, leader_response, leader):

        if leader_response and isinstance(leader_response, grpc.aio.AioRpcError) is False:
            
            await self.registry.register_leader(leader_response.host_address)
            await self.register_worker()

            id_not_set = await self.registry.leader_id_not_set(leader_response.host_address)
            if id_not_set:
                await self.registry.set_leader_id(leader_response.host_address, leader_response.host_id)

        else:
            await self.registry.remove_leader(leader.leader_address)




# import enum
# import statistics
# import multiprocessing
# import psutil
# import threading
# import asyncio
# import dill
# import time
# import ctypes
# from hedra.tools.helpers import awaitable
# from alive_progress import alive_bar
# from queue import Queue
# from hedra.command_line import CommandLine
# from hedra.reporting import Handler
# from .parallel.jobs import (
#     create_job,
#     run_job
# )


# class ParallelLocalWorker:

#     def __init__(self, config):
#         logger = Logger()
#         self.start = None
#         self.lock = threading.Lock()
#         self.session_logger = logger.generate_logger('hedra')
#         self.config = config
#         self.reporter_config = config.reporter_config
#         self.executor_config = config.executor_config
#         self._no_run_visuals = self.executor_config.get('no_run_visuals')
#         self._warmup_duration = self.executor_config.get('warmup', 0)
#         self._optimization_iters = self.executor_config.get('optimize', 0)
#         self._execute_warmup = self._warmup_duration > 0
#         self._execute_optimization = self._optimization_iters > 0
#         self._stages_completed = 0
#         self._monitor_thread = None

#         cpu_count = psutil.cpu_count()
#         self._size = self.executor_config.get(
#             'pool_size',
#             cpu_count
#         )
#         self.stage_queue = Queue(self._size)

#         if self._size < 1:
#             self.session_logger.warn(
#                 f'\nWarning - System either is single-core or requested less than one thread. Setting requested threads ot min of 1.\n'
#             )
#             self._size = 1

#         self._workers = self._size
#         self._stages = 6

#         if self._execute_warmup:
#             self._stages = 7

#         if self._execute_optimization:
#             self._stages = 8

#         self._barrier = multiprocessing.Barrier(self._workers)
#         self.stages = multiprocessing.Queue(self._stages * self._size)
#         self._current_stage = multiprocessing.Array(ctypes.c_char, 32)
#         self.optimized_aps = multiprocessing.Array(ctypes.c_float, self._workers)
#         self.optimized_batch_size = multiprocessing.Array(ctypes.c_int64, self._workers)
#         self.optimized_batch_time = multiprocessing.Array(ctypes.c_float, self._workers)
#         self._current_stage.value = b'initializing'

#         self._results = []
#         self._jobs = []
#         self._jobs_configs = []
#         self._total_actions = 0
#         self._progress_bar = None
#         self._visited = {
#             b'initializing': False,
#             b'setup': False,
#             b'warmup': False,
#             b'optimize': False,
#             b'execute': False,
#             b'results': False,
#             b'serializing': False
#         }

#     def _partition_actions(self, action=None):

#         configs = []
#         batch_size = self.config.executor_config.get('batch_size', 10**3)
#         worker_batch_size = int(batch_size/self._workers)
#         last_batch_size = worker_batch_size + (batch_size%self._workers)
#         last_batch_idx = self._workers - 1

#         with alive_bar(self._workers, title='Setting up workers...') as bar:
#             for idx, worker in enumerate(range(self._workers)):
#                 command_line = CommandLine()
#                 command_line.runner_mode = 'parallel'
#                 command_line = self.config.copy(command_line)
#                 command_line.executor_config.update(self.config.executor_config)
#                 command_line.reporter_config.update(self.config.reporter_config)
#                 command_line.log_level = self.config.log_level
#                 command_line.executor_config['pool_size'] = self._workers
#                 command_line.config_helper = None

#                 if command_line.executor_config.get('engine_type') == 'action-set':
#                     command_line.actions = {}
#                 else:
#                     command_line.actions = dict(self.config.actions)
                
#                 # command_line.executor_config['batch_size'] = batch_size
#                 if idx == last_batch_idx:
#                     command_line.executor_config['batch_size'] = last_batch_size
#                 else:
#                     command_line.executor_config['batch_size'] = worker_batch_size

#                 reporter_config = dict(self.reporter_config.copy())
#                 configs.append(dill.dumps({
#                     'config': command_line,
#                     'reporter_config': reporter_config,
#                     'worker_idx': worker
#                 }))

#                 bar()

#         return configs

#     def run(self):
#         self.session_logger.info(f'Starting parallel execution - {self._workers} - workers...\n')
#         configs = self._partition_actions(action='optimize')
#         self.session_logger.debug(f'Successfully generated bins, spawning threads...\n')
#         self.session_logger.info('\nStarting job...')

#         try:
#             self.start = time.time()
#             if self._no_run_visuals:
#                 self._run(configs)

#             else:
#                 with alive_bar(
#                     total=self._stages,
#                     title=f'Completed stages',
#                     bar=None, 
#                     spinner='dots_waves2',
#                     stats=False
#                 ) as bar:
#                     self._progress_bar = bar
#                     self._run(configs)
        
#         except Exception as err:
#             self.session_logger.error(f'\nErr - encountered exception while running parallel job - {str(err)}')
#             exit(1)


#     def _run(self, configs):
#         shared = (
#             self._barrier,
#             self.stages,
#             self._current_stage,
#             self.optimized_aps,
#             self.optimized_batch_size,
#             self.optimized_batch_time,
#         )

#         pool = multiprocessing.Pool(
#             processes=self._workers,
#             initializer=create_job,
#             initargs=shared
#         )
        
#         self._jobs = pool.map_async(run_job, configs)

#         if self._no_run_visuals:
            
#             self._results = self._jobs.get()

#         else:
#             self._stages_completed += 1
#             self._progress_bar()
#             self._progress_bar.text(f'- {self._current_stage.value.decode()}')
#             self._monitor()

#             self._results = self._jobs.get()

#             self._monitor_thread.join()

#     def _monitor(self):
#         self._monitor_thread = threading.Thread(target=self._monitor_async)
#         self._monitor_thread.start()

#     def _monitor_async(self):
#         loop = asyncio.new_event_loop()
#         asyncio.set_event_loop(loop)
#         loop.run_until_complete(self._monitor_progress())

#     async def _monitor_progress(self):
#         while self._stages_completed < self._stages:
#             if self._visited[self._current_stage.value] is False:
#                 self._visited[self._current_stage.value] = True
#                 await awaitable(self._progress_bar)

#                 current_stage = await awaitable(self._current_stage.value.decode)
#                 await awaitable(self._progress_bar.text, f'- {current_stage}')
#                 self._stages_completed += 1
#             await asyncio.sleep(1)
        
#         return 0
        
#     def kill(self):
#         loop = asyncio.get_event_loop()

#         handler = Handler(self.config)
#         loop.run_until_complete(handler.initialize_reporter())

#         actions_per_second_rates = []
#         total_completed_actions = []
#         elapsed_times = []
#         start_times = []
#         end_times = []

#         for result in self._results:
#             stats = result.get('stats')
#             actions_per_second_rates.append(stats.get('actions_per_second'))
#             total_completed_actions.append(stats.get('completed_actions'))
#             elapsed_times.append(stats.get('total_time'))
#             start_times.append(stats.get('start_time'))
#             end_times.append(stats.get('end_time'))

#         total_completed_actions = sum(total_completed_actions)
        
#         median_start = statistics.median(start_times)
#         median_end = statistics.median(end_times)
#         median_elapsed = median_end - median_start

#         median_actions_per_second = statistics.median(actions_per_second_rates)
#         peak_actions_per_second = sum(actions_per_second_rates)
#         average_actions_per_second = peak_actions_per_second/self._workers

#         self.session_logger.info(f'Calculated median APS per-worker of - {median_actions_per_second} - actions per second.')
#         self.session_logger.info(f'Calculated average APS per-worker of {average_actions_per_second} - actions per second.')
#         self.session_logger.info(f'Calculated estimated peak APS of - {peak_actions_per_second} - actions per second.')
#         self.session_logger.info(f'Total actions completed - {total_completed_actions} - over runtime of - {median_elapsed} - seconds')

#         self.session_logger.info(f'Submitting results for - {self._workers} - workers...')

#         for aggregated_events in self._results:
#             loop.run_until_complete(
#                 handler.merge(aggregated_events.get('events'))
#             )

#         stats = loop.run_until_complete(handler.get_stats())
#         loop.run_until_complete(handler.submit(stats))

#         self.session_logger.info('\nCompleted run, exiting...\n')



from enum import Enum


class PersonaTypes:
    BATCHED='BATCHED'
    DEFAULT='DEFAULT'
    RAMPED='RAMPED'
    RAMPED_INTERVAL='RAMPED_INTERVAL'
    CONSTANT_ARRIVAL='CONSTANT_ARRIVAL'
    CONSTANT_SPAWN='CONSTANT_SPAWN'
    SEQUENCE='SEQUENCE'
    WEIGHTED='WEIGHTED'
    NO_WAIT='NO-WAIT'


class PersonaTypesMap:

    def __init__(self) -> None:
        self.types = {
            'batched': PersonaTypes.BATCHED,
            'default': PersonaTypes.DEFAULT,
            'ramped': PersonaTypes.RAMPED,
            'constant-arrival': PersonaTypes.CONSTANT_ARRIVAL,
            'constant-spawn': PersonaTypes.CONSTANT_SPAWN,
            'sequence': PersonaTypes.SEQUENCE,
            'weighted': PersonaTypes.WEIGHTED,
            'no-wait': PersonaTypes.NO_WAIT
        }

    def __getitem__(self, persona_type: str):
        return self.types.get(persona_type, PersonaTypes.DEFAULT)


from typing import List


class HookMetadata:

    def __init__(
        self, 
        weight: int = 1, 
        order: int = 1, 
        env: str = None, 
        user: str = None, 
        tags: List[str] = []
    ) -> None:
        self.weight = weight
        self.order = order
        self.env = env
        self.user = user
        self.tags = tags


import imp


import uuid
from .service_node import ServiceNode


class RegisteredService:

    def __init__(self, expected_leaders) -> None:

        self.service_id = str(uuid.uuid4())
        self.registered_nodes = {}
        self.registered_count = 0
        self.requested_count = expected_leaders
        self.service_registered = False

    async def register_node(self, service_config):

        new_service_node = ServiceNode(service_config)

        is_registered = self.registered_nodes.get(new_service_node.service_address)

        if is_registered is None:
            self.registered_nodes[new_service_node.service_address] = new_service_node
            self.registered_count = len(self.registered_nodes)

            self.session_logger.info(f'Registered - {new_service_node.service_address} - for service - {new_service_node.service_name}.')

        else:
            self.session_logger.info(f'Re-registered - {new_service_node.service_address} - for service - {new_service_node.service_name}.')


    async def get_registration_status(self):
        self.registered_count = len(self.registered_nodes)
        
        return {
            'registered_addresses': list(self.registered_nodes.keys()),
            'registered_ids': [registered_node.service_id for registered_node in self.registered_nodes.values()],
            'registered_server_addresses': [registered_node.service_server_address for registered_node in self.registered_nodes.values()],
            'registered_count': self.registered_count,
            'ready': self.registered_count == self.requested_count
        }




from typing import Any, Optional, List


class SimpleContext:

    def __init__(self, **kwargs) -> None:
        for kwarg_name, kwarg in kwargs.items():
            object.__setattr__(self, kwarg_name, kwarg)

    def __str__(self) -> str:
        return str({
            key: value for key, value in self.__dict__.items() if key.startswith('__') is False
        })

    def __iter__(self):
        for key, value in self.__dict__.items():
            if key.startswith('__') is False:
                yield key, value

    def __getattribute__(self, __name: str) -> Any:
        return object.__getattribute__(self, __name)

    def __setattr__(self, __name: str, __value: Any) -> None:
        object.__setattr__(self, __name, __value)

    def __getitem__(self, name: str) -> Optional[Any]:
        if hasattr(self, name):
            return object.__getattribute__(self, name)

        return None

    def __setitem__(self, name: str, value: Any) -> None:
        object.__setattr__(self, name, value)

    def get(self, name: str) -> Optional[Any]:
        return self.__getitem__(name)       

    def keys(self) -> List[str]: 
        return [key for key in self.__dict__.keys() if key.startswith('__') is False]

    def valuses(self) -> List[Any]:
        return [value for key, value in self.__dict__.items() if key.startswith('__') is False]
    
    def items(self):
        return [
            (
                key, 
                value
            ) for key, value in self.__dict__.items() if key.startswith('__') is False
        ]

    def remove(self, name: str):
        if name.startswith('__') is False:
            object.__delattr__(self, name)
    


import os
import json
from flask import Flask
from multiprocessing import Process
from .services import UpdateService


class UpdateServer:

    def __init__(self, config=None):
        self.app = Flask(__name__)
        self.running = False
        self.update_service = UpdateService(config)
    
    def get_app(self):
        return self.app

    def run(self, host=None, port=6672):
        self.running = True
        self.app.run(host=host, port=port)

    def run_local(self, host=None, port=6672):
        self.running = True
        self.setup()
        self.app.run(host=host, port=port)

    def setup(self):
        for endpoint_config, endpoint_handler in self.update_service:   
            self.app.route(
                endpoint_config.get('path'),
                methods=endpoint_config.get('methods')
            )(endpoint_handler)
        



from typing import Dict
from pydantic import BaseModel
from hedra.reporting.types.common.types import ReporterTypes


class GraphiteConfig(BaseModel):
    host: str='localhost'
    port: int=2003
    custom_fields: Dict[str, str]={}
    reporter_type: ReporterTypes=ReporterTypes.Graphite


from hedra.core.engines.types.common.types import RequestTypes
from hedra.core.engines.types.http.result import HTTPResult
from .action import WebsocketAction


class WebsocketResult(HTTPResult):

    def __init__(self, action: WebsocketAction, error: Exception = None) -> None:
        super().__init__(action, error)      
        self.type = RequestTypes.WEBSOCKET


import psutil
from typing import List
from .time_parser import TimeParser


class Config:
    
    def __init__(self, **kwargs) -> None:

        for config_option_name, config_option_value in dict(kwargs).items():
            if config_option_value is None:
                del kwargs[config_option_name]
    
        

        self.total_time_string = kwargs.get('total_time', '1m')
        parsed_time = TimeParser(self.total_time_string)

        self.log_level = kwargs.get('log_level', 'info')
        self.persona_type = kwargs.get('persona_type', 'default')
        self.total_time = parsed_time.time
        self.batch_size = kwargs.get('batch_size', 1000)
        self.batch_interval = kwargs.get('batch_interval')
        self.action_interval = kwargs.get('action_interval', 0)
        self.optimize_iterations = kwargs.get('optimize_iterations', 0)
        self.optimizer_type = kwargs.get('optimizer_type', 'shg')
        self.batch_gradient = kwargs.get('batch_gradient', 0.1)
        self.cpus = kwargs.get('cpus', psutil.cpu_count(logical=False))
        self.no_run_visuals = kwargs.get('no_run_visuals', False)
        self.connect_timeout = kwargs.get('connect_timeout', 15)
        self.request_timeout = kwargs.get('request_timeout', 60)
        self.reset_connections = kwargs.get('reset_connections')
        self.graceful_stop = kwargs.get('graceful_stop', 1)
        self.optimized = False

        if self.request_timeout > self.total_time:
            self.request_timeout = self.total_time

        self.browser_type = kwargs.get('browser_type', 'chromium')
        self.device_type = kwargs.get('device_type')
        self.locale = kwargs.get('locale')
        self.geolocation = kwargs.get('geolocation')
        self.permissions: List[str] = kwargs.get('permissions', [])
        self.color_scheme = kwargs.get('color_scheme')
        self.group_size = kwargs.get('group_size')
        self.playwright_options = kwargs.get('playwright_options', {})



from hedra.core.graphs.stages import (
    Submit
)

from hedra.reporting.types import (
    StatsDConfig
)


class SubmitStatsDResultsStage(Submit):
    config=StatsDConfig(
        host='localhost',
        port=8125
    )


import asyncio
from collections import defaultdict
from hedra.core.engines.types.common.base_action import BaseAction
from hedra.core.engines.types import (
    MercuryGraphQLClient,
    MercuryGraphQLHTTP2Client,
    MercuryGRPCClient,
    MercuryHTTP2Client,
    MercuryHTTPClient,
    MercuryPlaywrightClient,
    MercuryWebsocketClient,
    MercuryUDPClient
)
from hedra.logging import HedraLogger
from typing import Any, Tuple, Union


class ActionsStore:

    def __init__(self, metadata_string: str) -> None:
        self.metadata_string = metadata_string
        self.actions = defaultdict(dict)
        self.sessions = defaultdict(dict)
        self._loop = None
        self.current_stage: str = None
        self.waiter = None
        self.setup_call = None
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

    def set_waiter(self, stage: str):

        if self._loop is None:
            self._loop = asyncio.get_event_loop()

        self.waiter = self._loop.create_future()
        self.current_stage = stage

    async def wait_for_ready(self, setup_call):
        await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Action Store waiting for Action or Task to notify store it is ready')
        self.setup_call = setup_call
        await self.waiter

        await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Action Store was notified and is exiting suspension')

    def store(self, request: str, action: Any, session: Any):

        self.actions[self.current_stage][request] = action
        self.sessions[self.current_stage][request] = session

        try:
            self.waiter.set_result(None)
        except asyncio.exceptions.CancelledError:
            pass

    def get(self, stage: str, action_name: str) -> Tuple[BaseAction, Union[MercuryGraphQLClient, MercuryGraphQLHTTP2Client, MercuryGRPCClient, MercuryHTTP2Client, MercuryHTTPClient, MercuryPlaywrightClient, MercuryWebsocketClient, MercuryUDPClient]]:
        action = self.actions.get(
            stage
        ).get(action_name)

        session = self.sessions.get(
            stage
        ).get(action_name)

        return action, session



import os
from pydantic import BaseModel
from hedra.reporting.types.common.types import ReporterTypes


class JSONConfig(BaseModel):
    events_filepath: str=f'{os.getcwd()}/events.json'
    metrics_filepath: str=f'{os.getcwd()}/metrics.json'
    reporter_type: ReporterTypes=ReporterTypes.JSON


import os
from setuptools import (
    setup,
    find_packages
)

current_directory = os.path.abspath(os.path.dirname(__file__))

with open(os.path.join(current_directory, 'README.md'), "r") as readme:
    package_description = readme.read()

version_string = ""
with open (os.path.join(current_directory, ".version"), 'r') as version_file:
    version_string = version_file.read()

setup(
    name="hedra",
    version=version_string,
    description="Performance testing at scale.",
    long_description=package_description,
    long_description_content_type="text/markdown",
    author="Sean Corbett",
    author_email="sean.corbett@umconnect.edu",
    url="https://github.com/scorbettUM/hedra",
    packages=find_packages(),
    keywords=[
        'pypi', 
        'cicd', 
        'python',
        'performance',
        'testing',
        'dag',
        'graph',
        'workflow'
    ],
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent"
    ],
    install_requires=[
        'aiologger',
        'yaspin',
        'attr',
        'networkx',
        'aiodns',
        'h2',
        'click',
        'psutil',
        'fastapi',
        'dill',
        'scipy',
        'art',
        'scikit-learn',
        'uvloop',
        'tdigest',
        'GitPython'
    ],
    entry_points = {
        'console_scripts': [
            'hedra=hedra.cli:run',
            'hedra-server=hedra.run_uwsgi:run_uwsgi'
        ],
    },
    extras_require = {
        'all': [
            'grpcio',
            'grpcio-tools',
            'gql',
            'playwright',
            'azure-cosmos',
            'libhoney',
            'influxdb_client',
            'newrelic',
            'aio_statsd',
            'prometheus-client',
            'prometheus-api-client',
            'cassandra-driver',
            'datadog',
            'motor',
            'redis',
            'aioredis',
            'aiomysql',
            'psycopg2-binary',
            'aiopg',
            'sqlalchemy',
            'boto3',
            'snowflake-connector-python',
            'google-cloud-bigquery',
            'google-cloud-bigtable',
            'google-cloud-storage',
        ],
        'all-engines': [
            'grpcio',
            'grpcio-tools',
            'gql',
            'playwright'

        ],
        'all-reporters': [
            'azure-cosmos',
            'libhoney',
            'influxdb_client',
            'newrelic',
            'aio_statsd',
            'prometheus-client',
            'prometheus-api-client',
            'cassandra-driver',
            'datadog',
            'motor',
            'redis',
            'aioredis',
            'aiomysql',
            'psycopg2-binary',
            'aiopg',
            'sqlalchemy',
            'boto3',
            'snowflake-connector-python',
            'google-cloud-bigquery',
            'google-cloud-bigtable',
            'google-cloud-storage',

        ],
        'playwright': [
            'playwright'
        ],
        'azure': [
            'azure-cosmos'
        ],
        'honeycomb': [
            'libhoney'
        ],
        'influxdb': [
            'influxdb_client'
        ],
        'newrelic': [
            'newrelic'
        ],
        'statsd': [
            'aio_statsd'
        ],
        'prometheus': [
            'prometheus-client',
            'prometheus-api-client'
        ],
        'cassandra': [
            'cassandra-driver'
        ],
        'datadog': [
            'datadog'
        ],
        'mongodb': [
            'motor'
        ],
        'redis': [
            'redis',
            'aioredis'
        ],
        'kafka': [
            'aiokafka'
        ],
        'sql': [
            'aiomysql',
            'psycopg2-binary',
            'aiopg',
            'sqlalchemy'
        ],
        'aws': [
            'boto3'
        ],
        'grpc': [
            'grpcio',
            'grpcio-tools'
        ],
        'graphql': [
            'gql'
        ],
        'snowflake': [
            'sqlalchemy',
            'snowflake-connector-python'
        ],
        'google': [
            'google-cloud-bigquery',
            'google-cloud-bigtable',
            'google-cloud-storage',
        ]
    },
    python_requires='>=3.8'
)



from .transitions import (
    submit_to_setup_transition,
    submit_to_optimize_transition,
    submit_to_execute_transition,
    submit_to_checkpoint_transition,
    submit_to_submit_transition,
    submit_to_complete_transition,
    submit_to_wait_transition
)


import uuid
from typing import List
from hedra.logging import HedraLogger
from hedra.reporting.events.types.base_event import BaseEvent
from hedra.reporting.metric import MetricsSet


try:
    from azure.cosmos.aio import CosmosClient
    from azure.cosmos import PartitionKey
    from .cosmosdb_config import CosmosDBConfig
    has_connector = True
except Exception:
    CosmosClient = None
    PartitionKey = None
    CosmosDBConfig = None
    has_connector = False


class CosmosDB:

    def __init__(self, config: CosmosDBConfig) -> None:
        self.account_uri = config.account_uri
        self.account_key = config.account_key

        self.database_name = config.database
        self.events_container_name = config.events_container
        self.metrics_container_name = config.metrics_container
        self.shared_metrics_container_name = f'{self.metrics_container_name}_metrics'
        self.errors_container_name = f'{self.metrics_container}_errors'
        self.events_partition_key = config.events_partition_key
        self.metrics_partition_key = config.metrics_partition_key
        self.custom_metrics_containers = {}

        self.analytics_ttl = config.analytics_ttl

        self.session_uuid = str(uuid.uuid4())
        self.metadata_string: str = None
        self.logger = HedraLogger()
        self.logger.initialize()

        self.events_container = None
        self.metrics_container = None
        self.shared_metrics_container = None
        self.errors_container = None
        self.client = None
        self.database = None

    async def connect(self):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connecting to CosmosDB')

        self.client = CosmosClient(
            self.account_uri,
            credential=self.account_key
        )


        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Connected to CosmosDB')
        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Database - {self.database_name} - if not exists')
        self.database = await self.client.create_database_if_not_exists(self.database_name)

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created or set Database - {self.database_name}')

    async def submit_events(self, events: List[BaseEvent]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Events container - {self.events_container_name} with Partition Key /{self.events_container_name} if not exists')
        self.events_container = await self.database.create_container_if_not_exists(
            self.events_container_name,
            PartitionKey(f'/{self.events_partition_key}')
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created or set Events container - {self.events_container_name} with Partition Key /{self.events_container_name}')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Events to container - {self.events_container_name} with Partition Key /{self.events_container_name}')
        for event in events:
            await self.events_container.upsert_item({
                'id': str(uuid.uuid4()),
                **event.record
            })

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Events to container - {self.events_container_name} with Partition Key /{self.events_container_name}')

    async def submit_common(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Shared Metrics container - {self.shared_metrics_container_name} with Partition Key /{self.shared_metrics_container_name} if not exists')
        self.shared_metrics_container = await self.database.create_container_if_not_exists(
            self.shared_metrics_container_name,
            PartitionKey(f'/{self.metrics_partition_key}')
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created or set Shared Metrics container - {self.shared_metrics_container_name} with Partition Key /{self.shared_metrics_container_name}')


        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Shared Metrics to container - {self.shared_metrics_container_name} with Partition Key /{self.shared_metrics_container_name}')
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Shared Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            await self.shared_metrics_container.upsert_item({
                'id': str(uuid.uuid4()),
                'name': metrics_set.name,
                'stage': metrics_set.stage,
                'group': 'common',
                **metrics_set.common_stats
            })

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Shared Metrics to container - {self.shared_metrics_container_name} with Partition Key /{self.shared_metrics_container_name}')
        
    async def submit_metrics(self, metrics: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Metrics container - {self.metrics_container_name} with Partition Key /{self.metrics_container_name} if not exists')
        self.metrics_container = await self.database.create_container_if_not_exists(
            self.metrics_container_name,
            PartitionKey(f'/{self.metrics_partition_key}')
        )

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created or set Metrics container - {self.metrics_container_name} with Partition Key /{self.metrics_container_name}')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Metrics to container - {self.metrics_container_name} with Partition Key /{self.metrics_container_name}')
        for metrics_set in metrics:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for group_name, group in metrics_set.groups.items():
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Metrics Group - {group_name}:{group.metrics_group_id}')

                await self.metrics_container.upsert_item({
                    'id': str(uuid.uuid4()),
                    'group': group_name,
                    **group.record
                })
            
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics to container - {self.metrics_container_name} with Partition Key /{self.metrics_container_name}')

    async def submit_custom(self, metrics_sets: List[MetricsSet]):
        
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for custom_group_name, group in metrics_set.custom_metrics.items():
                await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Custom Metrics Group - {custom_group_name}')

                custom_metrics_container_name = f'{custom_group_name}_metrics'
                await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Custom Metrics container - {custom_metrics_container_name} with Partition Key /{custom_metrics_container_name} if not exists')

                custom_container = await self.database.create_container_if_not_exists(
                    custom_metrics_container_name,
                    PartitionKey(f'/{self.metrics_partition_key}')
                )

                self.custom_metrics_containers[custom_metrics_container_name] = custom_container


                await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created or set Custom Metrics container - {custom_metrics_container_name} with Partition Key /{custom_metrics_container_name}')

                await custom_container.upsert_item({
                    'id': str(uuid.uuid4()),
                    'name': metrics_set.name,
                    'stage': metrics_set.stage,
                    'group': custom_group_name,
                    **group
                })

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Metrics to container - {self.metrics_container_name} with Partition Key /{self.metrics_container_name}')


    async def submit_errors(self, metrics_sets: List[MetricsSet]):

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Creating Error Metrics container - {self.errors_container_name} with Partition Key /{self.errors_container_name} if not exists')
        self.errors_container = await self.database.create_container_if_not_exists(
            self.errors_container_name,
            PartitionKey(f'/{self.metrics_partition_key}')
        )
        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Created or set Error Metrics container - {self.errors_container_name} with Partition Key /{self.errors_container_name}')

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitting Error Metrics to container - {self.errors_container_name} with Partition Key /{self.errors_container_name}')
        for metrics_set in metrics_sets:
            await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Submitting Error Metrics Set - {metrics_set.name}:{metrics_set.metrics_set_id}')

            for error in metrics_set.errors:
                await self.metrics_container.upsert_item({
                    'id': str(uuid.uuid4()),
                    'name': metrics_set.name,
                    'stage': metrics_set.stage,
                    'error_message': error.get('message'),
                    'error_count': error.get('count')
                })

        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Submitted Error Metrics to container - {self.errors_container_name} with Partition Key /{self.errors_container_name}')

    async def close(self):

        await self.logger.filesystem.aio['hedra.reporting'].debug(f'{self.metadata_string} - Closing session - {self.session_uuid}')
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closing connection to CosmosDB')

        await self.client.close()
        
        await self.logger.filesystem.aio['hedra.reporting'].info(f'{self.metadata_string} - Closed connection to CosmosDB')


import asyncio
import os
import json
import inspect
import uvloop
uvloop.install()
import sys
import importlib
import ntpath
from pathlib import Path
from hedra.core.graphs.stages.base.stage import Stage
from hedra.core.graphs import Graph
from hedra.logging import HedraLogger
from hedra.logging import (
    HedraLogger,
    LoggerTypes,
    logging_manager
)


def check_graph(path: str, log_level: str):

    logging_manager.disable(
        LoggerTypes.HEDRA, 
        LoggerTypes.DISTRIBUTED,
        LoggerTypes.FILESYSTEM,
        LoggerTypes.DISTRIBUTED_FILESYSTEM
    )
    logging_manager.update_log_level(log_level)

    logger = HedraLogger()
    logger.initialize()
    logging_manager.logfiles_directory = os.getcwd()

    graph_name = path
    if os.path.isfile(graph_name):
        graph_name = Path(graph_name).stem

    logger['console'].sync.info(f'Validating graph - {graph_name} - at - {path}.\n')

    hedra_config_filepath = os.path.join(
        os.getcwd(),
        '.hedra.json'
    )

    hedra_config = {}
    if os.path.exists(hedra_config_filepath):
        with open(hedra_config_filepath, 'r') as hedra_config_file:
            hedra_config = json.load(hedra_config_file)

    hedra_graphs = hedra_config.get('graphs', {})
    hedra_core_config = hedra_config.get('core', {
        'bypass_connection_validation': False,
        'connection_validation_retries': 3
    })
    
    if path in hedra_graphs:
        path = hedra_graphs.get(path)
    
    package_dir = Path(path).resolve().parent
    package_dir_path = str(package_dir)
    package_dir_module = package_dir_path.split('/')[-1]
    
    package = ntpath.basename(path)
    package_slug = package.split('.')[0]
    spec = importlib.util.spec_from_file_location(f'{package_dir_module}.{package_slug}', path)

    if path not in sys.path:
        sys.path.append(str(package_dir.parent))

    module = importlib.util.module_from_spec(spec)
    sys.modules[module.__name__] = module

    spec.loader.exec_module(module)

    logger['console'].sync.info('Loaded graph.')
    
    direct_decendants = list({cls.__name__: cls for cls in Stage.__subclasses__()}.values())

    discovered = {}
    for name, obj in inspect.getmembers(module):
        if inspect.isclass(obj) and issubclass(obj, Stage) and obj not in direct_decendants:
            discovered[name] = obj

    stages_count = len(discovered)

    logger['console'].sync.info(f'Validating - {stages_count} - stages.')

    loop = asyncio.new_event_loop()
    asyncio.set_event_loop(loop)

    graph = Graph(
        graph_name,
        list(discovered.values()),
        config={
            **hedra_core_config,
            'graph_module': module.__name__
        },
        cpus=1
    )

    graph.assemble()

    loop.run_until_complete(graph.check(path))

    logger['console'].sync.info('\nValidation complete!\n')

    os._exit(0)


from typing import Any, Dict
from pydantic import BaseModel
from typing import Optional
from hedra.reporting.types.common.types import ReporterTypes


class PrometheusConfig(BaseModel):
    pushgateway_address: str='localhost:9091'
    auth_request_method: str='GET'
    auth_request_timeout: int=60000
    auth_request_data: Dict[str, Any]={}
    username: Optional[str]
    password: Optional[str]
    namespace: Optional[str]
    job_name: str='hedra'
    custom_fields: Dict[str, str]={}
    reporter_type: ReporterTypes=ReporterTypes.Prometheus


from .registered_leader import RegisteredLeader


class NewLeaderQueue:

    def __init__(self) -> None:
        self.leaders = {}
        self.leader_addresses = []

    def __iter__(self):
        for leader in self.leader_addresses:
            yield leader

    async def __aiter__(self):
        for leader in self.leader_addresses:
            yield leader

    def __getitem__(self, leader_address):
        return self.leaders[leader_address]

    async def get_leaders(self):
        for leader_address in self.leader_addresses:
            yield self.leaders[leader_address]

    async def add(self, new_leader_request):
        leader_address = new_leader_request.host_address
        leader_id = new_leader_request.host_id
        self.leader_addresses += [leader_address]
        
        new_leader = RegisteredLeader(leader_address)
        new_leader.leader_id = leader_id

        self.leaders[leader_address] = new_leader

    async def clear(self):
        self.leader_addresses = []
        self.leaders = {}

    async def not_empty(self):
        return len(self.leaders) > 0



from .registered_worker import RegisteredWorker


class NewWorkerQueue:

    def __init__(self) -> None:
        self.workers = {}
        self.worker_addresses = []

    def __iter__(self):
        for worker in self.worker_addresses:
            yield worker

    async def __aiter__(self):
        for worker in self.worker_addresses:
            yield worker

    def __getitem__(self, worker_address):
        return self.workers[worker_address]

    async def get_leaders(self):
        for worker_address in self.worker_addresses:
            yield self.workers[worker_address]

    async def add(self, new_worker_request):
        worker_address = new_worker_request.host_address
        worker_id = new_worker_request.host_id
        self.worker_addresses += [worker_address]
        
        new_worker = RegisteredWorker(worker_address)
        new_worker.worker_id = worker_id

        self.workers[worker_address] = new_worker

    async def clear(self):
        self.worker_addresses = []
        self.workers = {}

    async def not_empty(self):
        return len(self.workers) > 0



from hedra.command_line import CommandLine


class ConfigStore:

    def __init__(self):
        self.configs = {}
        self.reporter_configs = {}

    def add_config(self, config_json):
        job_name = config_json.get('job_name')
        self._add_primary_config(job_name, config_json)

        return job_name

    def _add_primary_config(self, job_name, config_json):
        command_line = CommandLine()

        command_line.executor_config = config_json.get('executor_config')
        command_line.actions = config_json.get('actions')
        command_line.reporter_config = config_json.get('reporter_config', {})

        self.configs[job_name] = command_line

    def get(self, job_name):
        config = self.configs.get(job_name)
        reporter_config = self.reporter_configs.get(job_name)
        return config, reporter_config

    def delete_config(self, job):

        if self.configs.get(job):
            del self.configs[job]

        if self.reporter_configs.get(job):
            del self.reporter_configs[job]



import heapq

class Median:

    __slots__ = (
        'min_heap',
        'max_heap'
    )

    def __init__(self):
        self.min_heap = list()
        self.max_heap = list()

    def update(self, new_value):
        if not self.min_heap:
            heapq.heappush(self.min_heap, new_value)
        elif new_value >= self.min_heap[0]:
            heapq.heappush(self.min_heap, new_value)
        else:
            heapq.heappush(self.max_heap, -1 * new_value)            
            
        if len(self.min_heap) > len(self.max_heap) + 1:
            heapq.heappush(self.max_heap, -1 * heapq.heappop(self.min_heap))
        elif len(self.max_heap) > len(self.min_heap):
            heapq.heappush(self.min_heap, -1 * heapq.heappop(self.max_heap))
            
    def get(self):
        try:
            if len(self.min_heap) == len(self.max_heap):
                return 0.5 * (-1 * self.max_heap[0] + self.min_heap[0])
            return self.min_heap[0] 
        except IndexError:
            return 0



import asyncio
import time
import uuid
from typing import Awaitable, Dict
from hedra.core.engines.types.common.timeouts import Timeouts
from hedra.core.engines.types.common.concurrency import Semaphore
from hedra.core.engines.types.common.base_result import BaseResult
from .task import Task
from .result import TaskResult


class MercuryTaskRunner:

    __slots__ = (
        'session_id',
        'timeouts',
        'concurrency',
        'registered',
        'sem',
        'active',
        'waiter',
        'closed'
    )

    def __init__(self, concurrency: int=10**3, timeouts: Timeouts = Timeouts()) -> None:
        
        self.session_id = str(uuid.uuid4())

        self.timeouts = timeouts
        self.concurrency = concurrency

        self.registered: Dict[str, Task] = {}
        self.closed = False

        self.sem = asyncio.Semaphore(value=concurrency)
        self.active = 0
        self.waiter = None

    async def wait_for_active_threshold(self):
        if self.waiter is None:
            self.waiter = asyncio.get_event_loop().create_future()
            await self.waiter

    def extend_pool(self, increased_capacity: int):
        self.concurrency += increased_capacity
        
        self.sem = Semaphore(self.concurrency)

    def shrink_pool(self, decrease_capacity: int):
        self.concurrency -= decrease_capacity
        self.sem = Semaphore(self.concurrency)

    async def execute_prepared_request(self, task: Task) -> Awaitable[TaskResult]:

        result = None
        wait_start = time.monotonic()
        self.active += 1
 
        async with self.sem:
            
            try:

                if task.hooks.listen:
                    event = asyncio.Event()
                    task.hooks.channel_events.append(event)
                    await event.wait()

                start = time.monotonic()

                result: BaseResult = await asyncio.wait_for(
                    task.execute(),
                    timeout=self.timeouts.total_timeout
                )
                
                result.wait_start = wait_start
                result.start = start
                result.complete = time.monotonic()

                if task.hooks.notify:
                    await asyncio.gather(*[
                        asyncio.create_task(
                            channel(result, task.hooks.listeners)
                        ) for channel in task.hooks.channels
                    ])

                    for listener in task.hooks.listeners: 
                        if len(listener.hooks.channel_events) > 0:
                            event = listener.hooks.channel_events.pop()
                            if not event.is_set():
                                event.set()   

            except Exception as e:
                result = TaskResult(task)
                result.wait_start = wait_start
                result.start = start
                result.complete = time.monotonic()
                result.error = str(e)

            self.active -= 1
            if self.waiter and self.active <= self.concurrency:

                try:
                    self.waiter.set_result(None)
                    self.waiter = None

                except asyncio.InvalidStateError:
                    self.waiter = None

            return result

    async def close(self):
        pass



from .core.graphs.hooks import (
    action,
    after,
    before,
    channel,
    check,
    context,
    depends,
    event,
    setup,
    metric,
    restore,
    save,
    task,
    teardown,
    validate
)

from .core.graphs.stages import (
    Analyze,
    Checkpoint,
    Execute,
    Optimize,
    Setup,
    Teardown,
    Validate,
    Submit,
    Wait
)


from .reporting import (
    AWSLambdaConfig,
    AWSTimestreamConfig,
    BigQueryConfig,
    BigTableConfig,
    CassandraConfig,
    CloudwatchConfig,
    CosmosDBConfig,
    CSVConfig,
    DatadogConfig,
    DogStatsDConfig,
    GoogleCloudStorageConfig,
    GraphiteConfig,
    HoneycombConfig,
    InfluxDBConfig,
    JSONConfig,
    KafkaConfig,
    MongoDBConfig,
    MySQLConfig,
    NetdataConfig,
    NewRelicConfig,
    PostgresConfig,
    PrometheusConfig,
    RedisConfig,
    S3Config,
    SnowflakeConfig,
    SQLiteConfig,
    StatsDConfig,
    TelegrafConfig,
    TelegrafStatsDConfig,
    TimescaleDBConfig,
)


from .generated_setup_stage import SetupStage
from .generated_optimize_stage import OptimizeStage
from .generated_analyze_stage import AnalyzeStage
from .generated_checkpoint_stage import CheckpointStage
from .generated_teardown_stage import TeardownStage
from .generated_validate_stage import ValidateStage



import inspect
import os
from typing import Optional
from hedra.projects.generation import GraphGenerator
from hedra.core.graphs.stages.base.stage import Stage
from hedra.cli.exceptions.graph.create import InvalidStageType
from hedra.logging import HedraLogger
from hedra.logging import (
    HedraLogger,
    LoggerTypes,
    logging_manager
)


def create_graph(
    path: str, 
    stages: Optional[str], 
    engine: str,
    reporter: str,
    log_level: str
):

    logging_manager.disable(
        LoggerTypes.HEDRA, 
        LoggerTypes.DISTRIBUTED,
        LoggerTypes.FILESYSTEM,
        LoggerTypes.DISTRIBUTED_FILESYSTEM
    )
    logging_manager.update_log_level(log_level)

    logger = HedraLogger()
    logger.initialize()
    logging_manager.logfiles_directory = os.getcwd()

    logger['console'].sync.info(f'Creating new graph at - {path}.')

    if stages is None:
        stages_list = [
            'setup',
            'execute',
            'analyze',
            'submit'
        ]

    else:
        stages_list = stages.split(',')
    
    generated_stages_count = len(stages_list)
    generated_stages = ''.join([
        f'\n-{stage}' for stage in stages_list
    ])   

    logger['console'].sync.info(f'Generating - {generated_stages_count} stages:{generated_stages}') 


    generator = GraphGenerator()

    for stage in stages_list:
        if stage not in generator.valid_types:
            raise InvalidStageType(stage, [
                generator_type_name for generator_type_name, generator_type in generator.generator_types.items() if inspect.isclass(
                    generator_type
                ) and issubclass(generator_type, Stage)
            ])


    with open(path, 'w') as generated_test:
        generated_test.write(
            generator.generate_graph(
                stages_list,
                engine=engine,
                reporter=reporter
            )
        )

    logger['console'].sync.info('\nGraph generated!\n')


import os
from hedra.core.graphs.stages import (
    Submit
)

from hedra.reporting.types import (
    MongoDBConfig
)


class SubmitMongoDBResultsStage(Submit):
    config=MongoDBConfig(
        host='localhost:27017',
        username=os.getenv('MONGODB_USERNAME', ''),
        password=os.getenv('MONGODB_PASSWORD', ''),
        database='results',
        events_collection='events',
        metrics_collection='metrics'
    )


import asyncio
import time
import uuid
from typing import Union

from hedra.runners.leader_services.worker_registry import WorkerServicesManager
from hedra.runners.leader_services.pipelines import JobPipeline
from hedra.runners.leader_services.electorate import ElectionManager
from .job import Job


class JobRegistry:

    def __init__(self) -> None:

        logger = Logger()
        self.session_logger = logger.generate_logger()


        self.electorate = ElectionManager()

        self.results_task = None
        self.jobs = {}
        self.created = {}
        self.active = {}
        self.completed = {}
        self.failed = {}

        self.monitor = None
        self.ready = False
        self.monitoring = False

        self.leader_ip = None
        self.leader_port = None

        self.workers = WorkerServicesManager()
        
        self.leader_address = None
        self.leader_id = None

        logger = Logger()
        self.session_logger = logger.generate_logger()

        self.heartbeat_interval = None

    async def setup_leader_addresses(self, config):
        await self.electorate.setup_leader_addresses(config)

        self.leader_ip = self.electorate.leader_ip
        self.leader_port = self.electorate.leader_port
        self.leader_address = self.electorate.expected_leaders

    async def setup(self, config, leader_id):
        self.heartbeat_interval = config.distributed_config.get('heartbeat_interval', 1)

        await self.workers.setup(config, leader_id)
        await self.electorate.setup(config, leader_id)

        self.leader_ip = self.electorate.leader_ip
        self.leader_port = int(self.electorate.leader_port)
        self.leader_address = self.electorate.leader_address
        self.workers.leader_ip = self.electorate.leader_ip
        self.workers.leader_port = self.electorate.leader_port
        self.workers.leader_address = self.electorate.leader_address

    async def __iter__(self):
        for job_id in self.jobs:
            yield self.jobs[job_id]

    async def __aiter__(self):
        for job_id in self.jobs:
            yield self.jobs[job_id]

    def __getitem__(self, job_id):
        if job_id in self.jobs:
            return self.jobs[job_id]

        return None

    async def iter_created(self):
        for job_id in self.created:
            yield self.created[job_id]

    async def iter_active(self):
        for job_id in self.active:
            yield self.active[job_id]

    async def iter_completed(self):
        for job_id in self.completed:
            yield self.completed[job_id]

    async def iter_failed(self):
        for job_id in self.completed:
            yield self.completed[job_id]

    async def create_job(self, job_id, new_job, timeout=300):
        if job_id not in self.created and job_id not in self.active:

            self.session_logger.info(f'\nCreating new job id - {job_id}')

            pipeline = JobPipeline(
                job_id,
                leader_address=self.leader_address,
                leader_id=self.leader_id,
                timeout=timeout
            )

            workers = await self.workers.registered()
            await pipeline.setup_pipeline(
                new_job,
                workers
            )

            new_job = Job(job_id, pipeline, new_job)
            for worker in workers:
                await new_job.attach_worker_to_job(worker)

            self.created[job_id] = new_job  
            self.jobs[job_id] = new_job              

            await new_job.set_job_status('created')

    async def start_job(self, job_id):
        if job_id in self.created and job_id not in self.active:
            self.session_logger.info(f'Starting job - {job_id}.')

            job = self.created[job_id]

            await self.workers.create_new_worker_job(job)

            await job.start_timer()
            await job.set_job_status('active')

            self.active[job_id] = job
            
            del self.created[job_id]

    async def update_job_workers(self):
        workers = await self.workers.registered()
        for job_id in self.active:
            active_job = self.active[job_id]
            await active_job.pipeline.update_pipeline_workers(workers, self.workers.registry.failed)

    async def prune_expired_jobs(self):
        active = {}
        
        for job_id in self.active:
            active_job = self.active[job_id]
            await active_job.update_time_elapsed()
            expired = await active_job.check_if_deadline_exceeded()
            
            if expired:
                self.session_logger.info(f'Job - {active_job.job_id} - failed.')
                await active_job.set_job_status('failed')
                self.failed[job_id] = active_job
            else:
                active[job_id] = active_job

        self.active = active

    async def check_for_completed(self):
        active = {}

        for job_id in self.active:
            active_job = self.active[job_id]
            completed =await active_job.pipeline.check_if_completed()

            if completed:
                await active_job.set_job_status('completed')
                job_results = await self.workers.get_worker_job_results(active_job)
                active_job.results_task = asyncio.create_task(self.electorate.elect_leader_and_submit(job_results, job_id=active_job.job_id)) 

                self.completed[job_id] = active_job
            else:
                active[job_id] = active_job

        self.active = active

    async def register_leaders(self, config) -> Union[uuid.UUID, None]:
        return await self.electorate.register_leaders(config)

    async def update_leader_active_jobs(self):
        jobs = dict(self.jobs)
        for job_id in jobs:
            job = jobs[job_id]
            if job.status == 'created' or job.status == 'active':
                updates = await self.electorate.registered_services.share_job(job)

                for updated_pipeline_status in updates:
                    await jobs[job_id].pipeline.update_pipeline_status(updated_pipeline_status)

        self.jobs = jobs

    async def start_job_monitor(self):
        self.monitoring = True
        self.monitor = asyncio.create_task(
            self._monitor_jobs()
        )

    async def _monitor_jobs(self):
        while self.monitoring:
            await self.update_job_workers()
            await self.prune_expired_jobs()
            await asyncio.sleep(self.heartbeat_interval)

            for job_id in self.completed:
                completed_job = self.completed[job_id]
                if completed_job.results_task and completed_job.results_task.done():
                    await completed_job.results_task

            self.failed = {}

    async def stop_job_monitor(self):
        self.monitoring = False
        await self.monitor

    async def create_ephemeral_job(self, job_id, config):
        ephemeral_job = {
            'job_name': job_id,
            'executor_config': config.executor_config,
            'actions': config.actions
        }

        await self.create_job(
            job_id, 
            ephemeral_job, 
            config.distributed_config.get('job_timeout', 600)
        )

    async def wait_for_workers(self):
        start = time.time()
        elapsed = 0
        workers_registered = await self.workers.registry.registered_count_equals_expected_count()
        while workers_registered is False and elapsed < self.workers.worker_request_timeout:
            workers_registered = await self.workers.registry.registered_count_equals_expected_count()
            elapsed = time.time() - start
            await asyncio.sleep(1)

    async def wait_for_leaders(self):
        start = time.time()
        elapsed = 0
        leaders_registered = await self.electorate.registered_services.registered_count_equals_expected_count()
        while leaders_registered is False and elapsed < self.electorate.leader_election_timeout:
            leaders_registered = await self.electorate.registered_services.registered_count_equals_expected_count()
            elapsed = time.time() - start
            await asyncio.sleep(1)
        




import os
from pathlib import Path
from hedra.core.graphs.hooks.hook_types.hook_type import HookType
from hedra.core.graphs.hooks.registry.registry_types import RestoreHook
from hedra.core.graphs.stages.validate.exceptions import HookValidationError
from .base_hook_validator import BaseHookVaidator


class RestoreHookValidator(BaseHookVaidator):

    def __init__(self, metadata_string: str) -> None:
        super(
            RestoreHookValidator,
            self
        ).__init__(metadata_string)

    async def validate(self, hook: RestoreHook):

        try:

            await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Validating {hook.hook_type.name.capitalize()} Hook - {hook.name}:{hook.hook_id}:{hook.hook_id}')

            assert hook.hook_type is HookType.RESTORE, f"Hook type mismatch - hook {hook.name}:{hook.hook_id} is a {hook.hook_type.name} hook, but Hedra expected a {HookType.RESTORE.name} hook."
            assert hook.shortname in hook.name, f"Shortname {hook.shortname} must be contained in full Hook name {hook.name}:{hook.hook_id} for @restore hook {hook.name}:{hook.hook_id}."
            
            assert hook.call is not None, f"Method is not not found on stage or was not supplied to @restore hook - {hook.name}:{hook.hook_id}"
            assert 'self' in hook.call.__code__.co_varnames

            if hook.restore_path:
                hook_path_dir = str(Path(hook.restore_path).parent.resolve())
                
                assert hook.call.__code__.co_argcount > 1, f"Missing required argument 'data' for @restore hook {hook.name}:{hook.hook_id}"
                assert hook.call.__code__.co_argcount < 3, f"Too many args. - @restore hook {hook.name}:{hook.hook_id} only requires 'data' as additional args."
                assert hook.context_key is not None, f"Missing required keyword arg. - @restore hook {hook.name}:{hook.hook_id} requires a valid string for 'key'"
                assert hook.restore_path is not None, f"Missing required keyword arg. - @restore hook {hook.name}:{hook.hook_id} requires a valid string for 'checkpoint_filepath'"
                assert isinstance(hook.restore_path, str), f"Invalid path type - @restore hook {hook.name}:{hook.hook_id} path must be a valid string."
                assert os.path.exists(hook_path_dir), f"Invalid path - @restore hook {hook.name}:{hook.hook_id} path {hook_path_dir} must exist."
            
            await self.logger.filesystem.aio['hedra.core'].debug(f'{self.metadata_string} - Validated {hook.hook_type.name.capitalize()} Hook - {hook.name}:{hook.hook_id}:{hook.hook_id}')

        except AssertionError as hook_validation_error:
                raise HookValidationError(hook.stage_instance, str(hook_validation_error))


from hedra.core.graphs.stages.base.stage import Stage


class ReservedMethodError(Exception):

    def __init__(self, from_stage: Stage, reserve_method_name: str) -> None:
        self.from_stage = from_stage
        self.to_stage = None

        super().__init__(
            f'Stage Validation Error - Stage {from_stage.name} of type {from_stage.stage_type.name}:\nThe class method name - {reserve_method_name} - is reserved by Hedra. Please choose a different method name.'
        )
